// Vectors.

#ifndef _CL_V_H
#define _CL_V_H

#include "cl_object.h"

// A vector is a structure having the following interface:
//     v.length()        returns the number of elements
//     v[i]              returns the i-th element (0<=i<length), as a
//                       pseudo-lvalue (you can assign to it, but not take its
//                       address - exactly what you want for bit-vectors)
// This is implemented by letting v[i] be of a special "vector index" type.

template <class T> class cl_V;
template <class T> class cl_V_index;
template <class T> class cl_V_constindex;
template <class T> struct cl_V_vectorops;

template <class T>
class cl_V {
protected:
	uintL len; // number of elements
public:
	uintL length () const; // number of elements
	cl_V_vectorops<T>* vectorops; // get/set element
	const cl_V_index<T> operator[] (uintL index);
	const cl_V_constindex<T> operator[] (uintL index) const;
public: /* ugh */
	// Constructor.
	cl_V (uintL l, cl_V_vectorops<T>* ops) : len (l), vectorops (ops) {}
public:
	// Destructor.
	~cl_V ();
	// Ability to place an object at a given address.
	void* operator new (size_t size, cl_V* ptr) { (void)size; return ptr; }
private:
	void* operator new (size_t size, uintL l);
// No default constructor, copy constructor, assignment operator, new.
	cl_V ();
	cl_V (const cl_V&);
	cl_V& operator= (const cl_V&);
	void* operator new (size_t size);
};

template <class T>
class cl_V_index {
	// This is the class of objects created by accessing a non-const vector
	// through [].
public:
	cl_V<T>* vec;
	uintL index;
	operator T () const;
	void operator= (const T& x) const;
	// Constructor:
	cl_V_index (cl_V<T>* v, uintL i) : vec (v), index (i) {}
private:
// No default constructor, assignment operator.
	cl_V_index ();
	cl_V_index& operator= (const cl_V_index&);
};

template <class T>
class cl_V_constindex {
	// This is the class of objects created by accessing a const vector
	// through []. It lacks the assignment operator.
public:
	const cl_V<T>* vec;
	uintL index;
	operator T () const;
	// Constructor:
	cl_V_constindex (const cl_V<T>* v, uintL i) : vec (v), index (i) {}
private:
// No default constructor, assignment operator.
	cl_V_constindex ();
	cl_V_constindex& operator= (const cl_V_constindex&);
};

template <class T>
struct cl_V_vectorops {
	T (*element) (const cl_V<T>* vec, uintL index);
	void (*set_element) (cl_V<T>* vec, uintL index, const T& x);
	void (*do_delete) (cl_V<T>* vec);
};

// All member functions are inline.

template <class T>
inline uintL cl_V<T>::length () const
{
	return len;
}

template <class T>
inline const cl_V_index<T> cl_V<T>::operator[] (uintL index)
{
	return cl_V_index<T>(this,index);
}

template <class T>
inline const cl_V_constindex<T> cl_V<T>::operator[] (uintL index) const
{
	return cl_V_constindex<T>(this,index);
}

template <class T>
inline cl_V<T>::~cl_V ()
{
	vectorops->do_delete(this);
}

template <class T>
inline cl_V_index<T>::operator T () const
{
	return vec->vectorops->element(vec,index);
}

template <class T>
inline void cl_V_index<T>::operator= (const T& x) const
{
	vec->vectorops->set_element(vec,index,x);
}

template <class T>
inline cl_V_constindex<T>::operator T () const
{
	return vec->vectorops->element(vec,index);
}


#endif /* _CL_V_H */
